# Mutation Sentinel

Mutation Sentinel helps you *deeply* detect mutations on objects, arrays, and functions at runtime and enforce immutability in your codebase.

## Motivation

We wanted to optimize Flexport's React app with PureComponents. However, mutations can cause stale rendering bugs when used with PureComponents. We developed this library to help us incrementally detect and fix the mutations in our code, and enforce immutability along the way.

## Installation

```
npm install --save mutation-sentinel
```

## Usage

```js
import makeSentinel from "mutation-sentinel";

const obj = {};
const wrappedObj = makeSentinel(obj);
wrappedObj.value = "oops";
// console: Mutation detected by a sentinel!
```

This also works with arrays:

```js
const array = [];
const wrappedArray = makeSentinel(array);
wrappedArray.push("oops");
// console: Mutation detected by a sentinel!
```

And even deeply nested objects!

```js
const obj = {array: [{}]};
const wrappedObj = makeSentinel(obj);
wrappedObj.array[0].value = "oops";
// console: Mutation detected by a sentinel!
```

## Configuration

Your army of sentinels can be reconfigured globally at any time:

```js
import {configureSentinels} from "mutation-sentinel";

configureSentinels({
  allowMutationFor: obj => {
    // return true to NOT wrap obj with a sentinel
  },
  mutationHandler: mutation => {
    // respond to the mutation however you want
  }
});
```

The `mutation` object in `mutationHandler` has the following flow type:

```js
type Mutation =
  | {|
      type: "defineProperty",
      target: Observable,
      property: string,
      descriptor: Object,
    |}
  | {|
      type: "deleteProperty",
      target: Observable,
      property: string,
    |}
  | {|
      type: "set",
      target: Observable,
      property: string,
      value: any,
    |}
  | {|
      type: "setPrototypeOf",
      target: Observable,
      property: "[[Prototype]]",
      prototype: ?Object,
    |};

// Only Objects (includes Arrays) and Functions can be be wrapped.
type Observable = {} | (() => mixed);
```

## Browser Compatibility
This library relies on [the Proxy object](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Proxy). For browsers that do not support Proxies, `makeSentinel` simply returns the original object and no mutation detection occurs.

## Flexport's Use Case

Due to the size of Flexport’s app, we wanted the ability to purify our components incrementally, route by route. Since sentinels can be configured dynamically, we can enable and disable the `mutationHandler` on demand.

Here is the general approach that we took:
1. Wrap all of our flux store records with `makeSentinel`.
2. For a route we want to remove mutations for, configure the `mutationHandler` to log mutations to the console in development, and [Sentry](https://sentry.io) (our error reporting service) in production.
3. Deploy sentinels to production and fix the mutations as they are detected.
4. Once all the mutations are fixed, change `mutationHandler` to throw in development and no-op in production.

Our configuration looked something like:
https://gist.github.com/dounan/207cc05e47a97e22494739fcb42e2c3c

## Limitations
- Since the detection happens at runtime, we can’t find mutations in code that isn’t actually executed. We left the detection on in production for about a month to catch as many mutations as possible.
- It wasn’t feasible to wrap every object in our app with a sentinel, which means the unwrapped objects are still susceptible to undetected mutations.

## Gotchas
While a sentinel behaves the same as the original object, it is not equal to it.

```js
makeSentinel(myObj) !== myObj
```

Remember that shallow copies of sentinels are *not* themselves sentinels…but the nested objects of the shallow copy *are* sentinels.

```js
const obj = {nested: {}};
const wrappedObj = makeSentinel(obj);

// copiedObj is not a sentinel
const copiedObj = {...wrappedObj};
copiedObj.value = "oops"; // mutation is NOT detected

// copiedObj.nested is a sentinel because it was copied from wrappedObj
copiedObj.nested.value = "oops"; // MUTATION DETECTED!
```
